# -*- coding: utf-8 -*-

import urllib
import json
import urllib.request
import urllib.parse
import urllib.error
import ssl


class Http(object):
    def __init__(self, api_url: str, headers, json_load_switch=False, no_decode=False, json_parse=False):
        self.api_url = api_url
        self.headers = headers
        self.no_decode = no_decode
        self.json_load_switch = json_load_switch
        self.json_parse = json_parse

    def add_header(self, name: str, value: str):
        self.headers.append([name, value])

    def get_headers(self):
        return self.headers

    def delete_header(self, name: str):
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                if header[0] == name:
                    self.headers.remove(header)

    def post(self, uri: str, data=None, headers=None):
        url = self.api_url + uri
        request = urllib.request.Request(url, method="POST")
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                request.add_header(header[0], header[1])
        if headers is not None:
            for header in headers:
                request.add_header(header[0], header[1])
        return self.http_status(request, data)

    def get(self, uri: str, data=None, headers=None):
        if data is not None:
            url = self.api_url + uri + '?' + data
        else:
            url = self.api_url + uri
        request = urllib.request.Request(url, method="GET")
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                request.add_header(header[0], header[1])
        if headers is not None:
            for header in headers:
                request.add_header(header[0], header[1])
        return self.http_status(request)

    def delete(self, uri: str, data=None, headers=None):
        if data is not None:
            url = self.api_url + uri + '?' + data
        else:
            url = self.api_url + uri
        request = urllib.request.Request(url, method="DELETE")
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                request.add_header(header[0], header[1])
        if headers is not None:
            for header in headers:
                request.add_header(header[0], header[1])
        return self.http_status(request)

    def put(self, uri: str, data=None, headers=None):
        url = self.api_url + uri
        request = urllib.request.Request(url, method="PUT")
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                request.add_header(header[0], header[1])
        if headers is not None:
            for header in headers:
                request.add_header(header[0], header[1])
        return self.http_status(request, data)

    def patch(self, uri: str, data=None, headers=None):
        url = self.api_url + uri
        request = urllib.request.Request(url, method="PATCH")
        if self.get_headers().__len__() > 0:
            for header in self.get_headers():
                request.add_header(header[0], header[1])
        if headers is not None:
            for header in headers:
                request.add_header(header[0], header[1])
        return self.http_status(request, data)

    def http_status(self, request, data=None):
        try:
            context = ssl._create_unverified_context()
            if data is None:
                http_response = urllib.request.urlopen(request, context=context)
            else:
                if self.json_parse is True:
                    data = json.dumps(data)
                    data = bytes(data, encoding='utf8')
                    print('data=', data)
                    http_response = urllib.request.urlopen(request, data=data, context=context)
                else:
                    data = urllib.parse.urlencode(data)
                    data = bytes(data, encoding='utf8')
                    print('data=', data)
                    http_response = urllib.request.urlopen(request, data=data, context=context)
            if self.no_decode is True:
                return http_response.read()
            content = http_response.read().decode('utf-8')
            if self.json_load_switch is True:
                content = json.loads(content)
            return content
        except urllib.error.HTTPError as HTTPError:
            print('code:', HTTPError.code)
            print('reason:', HTTPError.reason)
            return HTTPError.code



# -*- coding: utf-8 -*-

import os
import tornado.ioloop
from cullinan.controller import handler_list, header_list
from dotenv import load_dotenv
from pathlib import Path
import tornado.ioloop
import tornado.options
import tornado.web
import tornado.httpserver
from tornado.options import define, options
import sys
import platform


def reflect(file: str, func: str):
    try:
        if func == 'nobody':
            __import__(file.replace('.py', ''))
        elif func == 'controller':
            __import__(file.replace('.py', ''))
        else:
            f = __import__(file.replace('.py', ''))
            function = getattr(f, func)
            function()
    except AttributeError:
        pass


def file_list_func():
    file_list = []
    for top, dirs, files in os.walk(os.getcwd()):
        for files_item in files:
            if os.path.splitext(files_item)[1] == '.py' and files_item != sys.argv[0][sys.argv[0].rfind(os.sep) + 1:]:
                if os.path.split(top)[-1] == os.path.split(os.getcwd())[-1]:
                    file_list.append(files_item)
                else:
                    system = platform.system()
                    if system == "Windows":
                        item = top.replace(os.path.split(os.path.realpath(sys.argv[0]))[0] + '\\', '')\
                                   .replace('\\', '.') + '.' + files_item
                        file_list.append(item)
                    else:
                        item = top.replace(os.path.split(os.path.realpath(sys.argv[0]))[0] + '/', '')\
                                   .replace('/', '.') + '.' + files_item
                        file_list.append(item)
    return file_list


def scan_controller(file_path: list):
    for x in file_path:
        reflect(x, 'controller')


def scan_service(file_path: list):
    for x in file_path:
        reflect(x, 'nobody')


def get_index_list(url_list: list) -> list:
    index_list = []
    for index in range(0, url_list.__len__()):
        if url_list[index] == '([a-zA-Z0-9-]+)':
            index_list.append(index)
    index_list.append('*')
    return index_list


def sort_url():
    handler_list_length = []
    for index in range(0, handler_list.__len__()):
        handler_list[index] = list(handler_list[index])
        handler_list[index][0] = handler_list[index][0].split('/')
        index_list = get_index_list(handler_list[index][0])
        handler_list_length.append(index_list.__len__())
        handler_list[index].append(index_list)
    for index in range(0, max(handler_list_length)):
        for i in range(0, handler_list.__len__()):
            for j in range(i + 1, handler_list.__len__()):
                if handler_list[i][2].__len__() >= index + 1 and handler_list[j][2].__len__() >= index + 1:
                    if handler_list[i][2][index] != '*' and handler_list[j][2][index] != '*':
                        if handler_list[i][2][index] < handler_list[j][2][index]:
                            handler_list[i], handler_list[j] = handler_list[j], handler_list[i]
                    elif handler_list[i][2][index] != '*' and handler_list[j][2][index] == '*':
                        handler_list[i], handler_list[j] = handler_list[j], handler_list[i]
                    elif handler_list[i][2][index] == '*':
                        continue
    for item in handler_list:
        url = ""
        for index in range(1, len(item[0])):
            url = url + "/" + item[0][index]
        item[0] = url
        del item[2]


def run(handlers=None):
    if handlers is None:
        handlers = []
    print(
        "\n|||||||||||||||||||||||||||||||||||||||||||||||||\n|||                                           |||\n|||  "
        "   _____      _ _ _ "
        "               |||\n|||    / ____|    | | (_)                     |||\n|||   | |    _   _| | |_ _ __   __ "
        "_ _ __     |||\n|||   | |   | | | | | | | '_ \ / _` | '_ \    |||\n|||   | |___| |_| | | | | | | | (_| | | | "
        "|   |||\n|||    \_____\__, "
        "_|_|_|_|_| |_|\__,_|_| |_|   |||\n|||                                           "
        "|||\n|||||||||||||||||||||||||||||||||||||||||||||||||\n\t|||")
    print("\t|||\tloading env...")
    load_dotenv()
    load_dotenv(verbose=True)
    env_path = Path(os.getcwd()) / '.env'
    load_dotenv(dotenv_path=env_path)
    settings = dict(
        template_path=os.path.join(os.getcwd(), 'templates'),
        static_path=os.path.join(os.getcwd(), 'static')
    )
    print("\t|||\t\t└---scanning controller...")
    print("\t|||\t\t\t...")
    scan_service(file_list_func())
    scan_controller(file_list_func())
    sort_url()
    mapping = tornado.web.Application(
        handlers=handler_list + handlers,
        **settings
    )
    print("\t|||\t\t└---loading controller finish\n\t|||\t")
    define("port", default=os.getenv("SERVER_PORT"), help="run on the given port", type=int)
    print("\t|||\tloading env finish\n\t|||\t")
    http_server = tornado.httpserver.HTTPServer(mapping)
    if os.getenv("SERVER_THREAD") is not None:
        print("\t|||\t\033[0;36;0mserver is starting \033[0m")
        print("\t|||\t\033[0;36;0mport is " + str(os.getenv("SERVER_PORT")) + " \033[0m")
        http_server.bind(options.port)
        http_server.start(int(os.getenv("SERVER_THREAD")) | 0)
    else:
        http_server.listen(options.port)
        print("\t|||\t\033[0;36;0mserver is starting \033[0m")
        print("\t|||\t\033[0;36;0mport is " + str(os.getenv("SERVER_PORT")) + " \033[0m")
    tornado.ioloop.IOLoop.current().start()


def add_global_header(name: str, value: str):
    header_list.append([name, value])



# -*- coding: utf-8 -*-

import json
import tornado.web
import tornado.websocket
import types
import functools
from cullinan.service import service_list, response_build
from typing import Callable

handler_list = []
header_list = []
controller_func_list = []
KEY_NAME_INDEX = {
    "url_params": 0,
    "query_params": 1,
    "body_params": 2,
    "file_params": 3,
}


class EncapsulationHandler(object):
    @classmethod
    def set_fragment_method(cls, cls_name: str, func: Callable[[object, tuple, dict], None]):
        @functools.wraps(func)
        def dummy(self, *args, **kwargs):
            func(self, *args, **kwargs)

        setattr(cls_name, func.__name__, dummy)

    @staticmethod
    def add_func(**kwargs: dict) -> Callable:
        def inner(func: Callable):
            controller_func_list.append((kwargs['url'], func, kwargs['type']))

        return inner

    @staticmethod
    def add_url(url: str, f: Callable) -> object:
        servlet = type('Servlet' + url.replace('/', ''), (Handler,),
                       {"set_instance_method": EncapsulationHandler.set_fragment_method})
        if handler_list.__len__() == 0:
            servlet.set_instance_method(servlet, f)
            servlet.f = types.MethodType(f, servlet)
            handler_list.append((url, servlet))
            return servlet
        else:
            for item in handler_list:
                if url == item[0]:
                    item[1].set_instance_method(item[1], f)
                    item[1].f = types.MethodType(f, item[1])
                    return item[1]
            else:
                servlet.set_instance_method(servlet, f)
                servlet.f = types.MethodType(f, servlet)
                handler_list.append((url, servlet))
                return servlet

    @staticmethod
    def add_url_ws(url: str, cls: Callable) -> object:
        servlet = type('Servlet' + url.replace('/', ''), (tornado.websocket.WebSocketHandler,),
                       {"set_instance_method": EncapsulationHandler.set_fragment_method})
        setattr(servlet, 'service', service_list)
        if handler_list.__len__() == 0:
            for item in dir(cls):
                if not item.startswith('__') and not item.endswith('__'):
                    setattr(servlet, item, cls.__dict__[item])
            handler_list.append((url, servlet))
            return servlet
        else:
            for item in handler_list:
                if url == item[0]:
                    for i in dir(cls):
                        if not i.startswith('__') and not i.endswith('__'):
                            setattr(servlet, item, cls.__dict__[item])
                    return item[1]
            else:
                for item in dir(cls):
                    if not item.startswith('__') and not item.endswith('__'):
                        setattr(servlet, item, cls.__dict__[item])
                handler_list.append((url, servlet))
                return servlet


class Handler(tornado.web.RequestHandler):

    def data_received(self, chunk):
        pass

    def set_default_headers(self):
        if header_list.__len__() > 0:
            for header in header_list:
                self.set_header(header[0], header[1])

    def options(self):
        pass


def request_resolver(self, url_param_key_list: tuple, url_param_value_list: tuple,
                     query_param_names: tuple, body_param_names, file_param_key_list) -> tuple:
    if url_param_key_list.__len__() > 0 and query_param_names is not None and body_param_names is not None and\
            file_param_key_list is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        query_param_dict = {}
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        body_param_dict = {}
        file_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
            for name in file_param_key_list:
                file_param_dict[name] = self.request.files[name]
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return url_param_dict, query_param_dict, body_param_dict, file_param_dict
    if url_param_key_list.__len__() > 0 and query_param_names is not None and body_param_names is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        query_param_dict = {}
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        body_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        return url_param_dict, query_param_dict, body_param_dict
    if url_param_key_list.__len__() > 0 and query_param_names is not None and file_param_key_list is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        query_param_dict = {}
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        file_param_dict = {}
        for name in file_param_key_list:
            file_param_dict[name] = self.request.files[name]
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return url_param_dict, query_param_dict, None, file_param_dict
    elif query_param_names is not None and url_param_key_list is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        query_param_dict = dict()
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        return url_param_dict, query_param_dict, None, None
    elif url_param_key_list is not None and file_param_key_list is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        file_param_dict = {}
        for name in file_param_key_list:
            file_param_dict[name] = self.request.files[name]
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return url_param_dict, None, None, file_param_dict
    elif file_param_key_list is not None and body_param_names is not None:
        file_param_dict = {}
        body_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
            for name in file_param_key_list:
                body_param_dict[name] = self.request.files[name]
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return None, None, body_param_dict, file_param_dict
    elif query_param_names is not None and file_param_key_list is not None:
        query_param_dict = {}
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        file_param_dict = {}
        for name in file_param_key_list:
            file_param_dict[name] = self.request.files[name]
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return None, query_param_dict, None, file_param_dict
    elif url_param_key_list is not None and body_param_names is not None:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        body_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        return url_param_dict, None, body_param_dict, None
    elif query_param_names is not None and body_param_names is not None:
        query_param_dict = {}
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        body_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        return None, query_param_dict, body_param_dict, None
    elif url_param_key_list.__len__() > 0:
        url_param_dict = {}
        for index in range(0, url_param_key_list.__len__()):
            url_param_dict[url_param_key_list[index]] = url_param_value_list[index]
        print("\t|||\t url_params", end="")
        print(url_param_dict)
        return url_param_dict, None, None, None
    elif query_param_names is not None:
        query_param_dict = dict()
        for name in query_param_names:
            query_param_dict[name] = self.get_query_argument(name)
        print("\t|||\t query_params", end="")
        print(query_param_dict)
        return None, query_param_dict, None, None
    elif file_param_key_list is not None:
        file_param_dict = {}
        for name in file_param_key_list:
            file_param_dict[name] = self.request.files[name]
        print("\t|||\t file_params", end="")
        print(file_param_dict.keys())
        return None, None, None, file_param_dict
    elif body_param_names is not None:
        body_param_dict = {}
        if self.request.headers.get("Content-Type") == 'application/json':
            json_data = self.request.body
            data = json.loads(json_data)
            for name in body_param_names:
                body_param_dict[name] = data[name]
        else:
            for name in body_param_names:
                body_param_dict[name] = self.get_body_argument(name)
        print("\t|||\t body_params", end="")
        print(body_param_dict)
        return None, None, body_param_dict, None
    else:
        return None, None, None, None


def header_resolver(self, header_names: list):
    if header_names is not None:
        need_header = dict()
        for name in header_names:
            need_header[name] = self.request.headers.get(name, None)
            if need_header[name] is not None:
                print("\t|||\t request_headers", end="")
                print(need_header)
            else:
                # TODO error
                raise Exception("missing header")
        return need_header
    else:
        return None


def url_resolver(url: str) -> tuple:
    find_all = lambda origin, target: [i for i in range(origin.find(target), len(origin)) if origin[i] == target]
    before_list = find_all(url, "{")
    after_list = find_all(url, "}")
    url_param_list = []
    for index in range(0, before_list.__len__()):
        url_param_list.append(url[int(before_list[index]) + 1:int(after_list[index])])
    for url_param in url_param_list:
        url = url.replace(url_param, "[a-zA-Z0-9-]+")
    url = url.replace("{", "(").replace("}", ")/*")
    return url, url_param_list


def request_handler(self, func, params, headers, type, get_request_body=False):
    global response
    global controller_self
    if type == 'get':
        controller_self = self.get_controller_self
    elif type == 'post':
        controller_self = self.post_controller_self
    elif type == 'patch':
        controller_self = self.patch_controller_self
    elif type == 'delete':
        controller_self = self.delete_controller_self
    elif type == 'put':
        controller_self = self.put_controller_self
    setattr(controller_self, 'service', service_list)
    setattr(controller_self, 'response', response_build)
    param_names = func.__code__.co_varnames
    param_names = list(param_names)
    if "self" in param_names:
        param_names.remove("self")
    else:
        raise Exception("controller's params must have self")
    if len(param_names) == 0:
        response = func(controller_self)
    else:
        param_list = []
        for item in param_names:
            if item in KEY_NAME_INDEX.keys():
                if KEY_NAME_INDEX[item] is not None:
                    param_list.append(params[KEY_NAME_INDEX[item]])
            elif item == 'request_body' and get_request_body is True:
                param_list.append(self.request.body)
            elif item == 'headers' and headers is not None:
                param_list.append(headers)
        if len(param_list) == 0:
            response = func(controller_self)
        elif len(param_list) == 1:
            response = func(controller_self, param_list[0])
        elif len(param_list) == 2:
            response = func(controller_self, param_list[0], param_list[1])
        elif len(param_list) == 3:
            response = func(controller_self, param_list[0], param_list[1], param_list[2])
        elif len(param_list) == 4:
            response = func(controller_self, param_list[0], param_list[1], param_list[2],
                            param_list[3])
        elif len(param_list) == 5:
            response = func(controller_self, param_list[0], param_list[1], param_list[2],
                            param_list[3], param_list[4])
        elif len(param_list) == 6:
            response = func(controller_self, param_list[0], param_list[1], param_list[2],
                            param_list[3], param_list[4], param_list[5])
    if header_list.__len__() > 0:
        for header in header_list:
            self.set_header(header[0], header[1])
    if response.get_headers().__len__() > 0:
        for header in response.get_headers():
            self.set_header(header[0], header[1])
    self.set_status(response.get_status())
    self.write(response.get_body())
    response.__body__ = ''
    response.__headers__ = []
    response.__status__ = 200
    response.__status_msg__ = ''
    response.__is_static__ = False
    self.finish()


def get_api(**kwargs):
    def inner(func):
        url_param_key_list = []
        if kwargs['url'].find("{") != -1:
            kwargs['url'], url_param_key_list = url_resolver(kwargs['url'])

        @EncapsulationHandler.add_func(url=kwargs['url'], type='get')
        def get(self, *args):
            print("\t||| request:")
            if self.get_controller_url_param_key_list is not None:
                request_handler(self,
                                func,
                                request_resolver(self, self.get_controller_url_param_key_list + url_param_key_list,
                                                 args,
                                                 kwargs.get('query_params', None), None,
                                                 kwargs.get('file_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'get')
            else:
                request_handler(self,
                                func,
                                request_resolver(self, url_param_key_list,
                                                 args,
                                                 kwargs.get('query_params', None), None,
                                                 kwargs.get('file_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'get')

        return get

    return inner


def post_api(**kwargs):
    def inner(func):
        url_param_key_list = []
        if kwargs['url'].find("{") != -1:
            kwargs['url'], url_param_key_list = url_resolver(kwargs['url'])

        @EncapsulationHandler.add_func(url=kwargs['url'], type='post')
        def post(self, *args):
            print("\t||| request:")
            if self.post_controller_url_param_key_list is not None:
                request_handler(self,
                                func,
                                request_resolver(self, self.post_controller_url_param_key_list + url_param_key_list,
                                                 args,
                                                 kwargs.get('query_params', None),
                                                 kwargs.get('body_params', None),
                                                 kwargs.get('file_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'post',
                                kwargs.get('get_request_body', False))
            else:
                request_handler(self,
                                func,
                                request_resolver(self, url_param_key_list,
                                                 args,
                                                 kwargs.get('query_params', None),
                                                 kwargs.get('body_params', None),
                                                 kwargs.get('file_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'post',
                                kwargs.get('get_request_body', False))

        return post

    return inner


def patch_api(**kwargs):
    def inner(func):
        url_param_key_list = []
        if kwargs['url'].find("{") != -1:
            kwargs['url'], url_param_key_list = url_resolver(kwargs['url'])

        @EncapsulationHandler.add_func(url=kwargs['url'], type='patch')
        def patch(self, *args):
            print("\t||| request:")
            if self.patch_controller_url_param_key_list is not None:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 self.patch_controller_url_param_key_list + url_param_key_list, args,
                                                 kwargs.get('query_params', None),
                                                 kwargs.get('body_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'patch',
                                kwargs.get('get_request_body', False))
            else:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 url_param_key_list, args,
                                                 kwargs.get('query_params', None),
                                                 kwargs.get('body_params', None)),
                                header_resolver(self, kwargs.get('headers', None)),
                                'patch',
                                kwargs.get('get_request_body', False))

        return patch

    return inner


def delete_api(**kwargs):
    def inner(func):
        url_param_key_list = []
        if kwargs['url'].find("{") != -1:
            kwargs['url'], url_param_key_list = url_resolver(kwargs['url'])

        @EncapsulationHandler.add_func(url=kwargs['url'], type='delete')
        def delete(self, *args):
            print("\t||| request:")
            if self.delete_controller_url_param_key_list is not None:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 self.delete_controller_url_param_key_list + url_param_key_list, args,
                                                 kwargs.get('query_params', None), None),
                                header_resolver(self, kwargs.get('headers', None)),
                                'delete')
            else:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 url_param_key_list, args,
                                                 kwargs.get('query_params', None), None),
                                header_resolver(self, kwargs.get('headers', None)),
                                'delete')

        return delete

    return inner


def put_api(**kwargs):
    def inner(func):
        url_param_key_list = []
        if kwargs['url'].find("{") != -1:
            kwargs['url'], url_param_key_list = url_resolver(kwargs['url'])

        @EncapsulationHandler.add_func(url=kwargs['url'], type='put')
        def put(self, *args):
            print("\t||| request:")
            if self.put_controller_url_param_key_list is not None:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 self.put_controller_url_param_key_list + url_param_key_list, args,
                                                 kwargs.get('query_params', None), None),
                                header_resolver(self, kwargs.get('headers', None)),
                                'put')
            else:
                request_handler(self,
                                func,
                                request_resolver(self,
                                                 url_param_key_list, args,
                                                 kwargs.get('query_params', None), None),
                                header_resolver(self, kwargs.get('headers', None)),
                                'put')

        return put

    return inner


def controller(**kwargs) -> Callable:
    url = kwargs.get('url', '')
    global url_params
    url_params = None
    if url != '':
        url, url_params = url_resolver(url)

    def inner(cls):
        for item in controller_func_list:
            if controller_func_list.__len__() != 0:
                handler = EncapsulationHandler.add_url(url + item[0], item[1])
                setattr(handler, item[2] + '_controller_self', cls)
                setattr(handler, item[2] + '_controller_url_param_key_list', url_params)
        controller_func_list.clear()

    return inner



# -*- coding: utf-8 -*-

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import *
from sqlalchemy.dialects.mysql import *
import os


class Conn(object):
    db_url = 'mysql+pymysql://{}:{}@{}:{}/{}?charset={}'.format(
        os.getenv("DB_USERNAME"),
        os.getenv("DB_PASSWORD"),
        os.getenv("DB_HOST"),
        os.getenv("DB_PORT"),
        os.getenv("DB_NAME"),
        os.getenv("DB_CODING"),
        os.getenv("DB_POOL_RECYCLE"),
        os.getenv("DB_POOL_SIZE"),
        os.getenv("DB_MAX_OVERFLOW")
    )
    engine = create_engine(db_url)
    Base = declarative_base(engine)

    @classmethod
    def conn(cls):
        session = sessionmaker(Conn.engine)
        session = session()
        return session

    @classmethod
    def save(cls):
        Conn.Base.metadata.create_all(Conn.engine)

